import logging
import sys
import time
import threading


from queue import Queue
from .worker import ledWorker
from .methods import *  # NOQA

if sys.platform == "linux":
    import RPi.GPIO as GPIO


class ledPulse:
    """
    There are 4 states of pulsing:
     - on: steady light at value;
     - off: turn off led;
     - normal: pulse with delay;
     - fast: pulse with delay/2;
     """

    delayMethods = ['constant', 'sin', 'cos']
    valueMethods = ['on', 'off', 'linear', 'sin', 'cos']
    methods = {'delay': {}, 'value': {}}

    defaultSet = {
        'min': 2,
        'max': 100,
        'delay': 0.01,
        'delayMethod': 'constant',
        'initialMethod': None,
        'loopMethod': 'linear',
        'finalMethod': None
    }

    def __init__(self, gpio, name="led0"):
        self.log = logging.getLogger(self.__class__.__name__)
        self.gpio = gpio
        self.name = name

        self.changed = threading.Event()

        if sys.platform == "linux":
            GPIO.setwarnings(False)       # disable warnings
            GPIO.setmode(GPIO.BCM)      # set pin numbering system
            # set GPIO for output
            GPIO.setup(self.gpio, GPIO.OUT)

            # create PWM instance with frequency
            self.pwm = GPIO.PWM(self.gpio, 120)
            # self.setValue = self.pwm.ChangeDutyCycle
            self.supported = True
            # self.pwm.start(0)  # start pwm with value 0
        else:
            # Not supported, skip setValue
            # self.setValue = print
            self.supported = False
            # self.setValue = lambda x: x

        self.log.info(f"platform '{sys.platform}' " +
                      f"support: {sys.platform == 'linux'}")

        for dm in self.delayMethods:
            self.register_delay_method(dm, eval(f"delay_{dm}"))

        for vm in self.valueMethods:
            self.register_value_method(vm, eval(f"value_{vm}"))

        # Add thread placeholder
        self.worker = ledWorker(self)
        self.queue = Queue()

        # Set default values
        for key, value in self.defaultSet.items():
            setattr(self, key, value)

    def setValue(self, value):
        if self.supported:
            self.pwm.ChangeDutyCycle(value)
        else:
            # on unsupported platforms, we just print the value
            print(value)

    def set(self, **kwargs):

        # for key, value in self.defaultSet.items():
        #     setattr(self, key, value)

        for key, value in kwargs.items():
            if key in self.defaultSet.keys():
                setattr(self, key, value)
                self._set = True
            else:
                self.log.warning(f"set: unknown parameter '{key}'")

        self.log.debug(f"fill values table, "
                       f"loopMethod:{self.loopMethod}, "
                       f"delayMethod:{self.delayMethod}, min:{self.min}, "
                       f"max:{self.max}, delay:{self.delay}")

        delayValues = []
        for step in range(0, 100):
            try:
                delay = self.methods['delay'][self.delayMethod](self, step)
            except NameError:
                self.log.error(f"delay method '{self.delayMethod}' not found")
                return
            delayValues.append(delay)

        initialValues = []
        if self.initialMethod:
            for step in range(0, 50):
                try:
                    value = self.methods['value'][self.initialMethod](self, step)
                except KeyError:
                    self.log.error(f"value method '{self.initialMethod}' not found")
                    return
                initialValues.append((value, delayValues[step]))

        loopValues = []
        for step in range(0, 100):
            try:
                value = self.methods['value'][self.loopMethod](self, step)
            except KeyError:
                self.log.error(f"value method '{self.loopMethod}' not found")
                return
            loopValues.append((value, delayValues[step]))

        finalValues = []
        if self.finalMethod:
            for step in range(50, 100):
                try:
                    value = self.methods['value'][self.finalMethod](self, step)
                except KeyError:
                    self.log.error(f"value method '{self.finalMethod}' not found")
                    return
                finalValues.append((value, delayValues[step]))

        self.queue.put((initialValues, loopValues, finalValues))

    def start(self):
        if self.supported:
            self.pwm.start(0)
        self.worker = ledWorker(self)
        self.worker.start()

    def stop(self):
        self.queue.put(([], [], []))
        self.worker.join()
        if self.supported:
            self.pwm.stop()
        self.worker = None


    ###
    # PLUGINS METHODS
    ###
    def register_value_method(self, name, fn):
        self.register_method('value', name, fn)

    def register_delay_method(self, name, fn):
        self.register_method('delay', name, fn)

    def register_method(self, kind, name, fn):
        if kind not in ['delay', 'value']:
            self.log.warning(f"unknown kind {kind}")
            return

        if name not in self.methods[kind]:
            self.methods[kind][name] = fn
            self.log.info(f"registered {kind} method '{name}'")
        else:
            self.log.warning(f"{kind} method '{name}' already defined")
